#include <cstdio>
#include <cstring>

using namespace std;

#include "bitfun.h"
#include "bitbuf.hh"

static int verbose = 4;

void BitBuf::Resize(unsigned newbitsize)
{
   	unsigned need = (newbitsize+7) >> 3;
   	if(alloced >= need)return;
   	
   	alloced = (need+0xFFF) & ~0xFFF;
   	
   	ptr.resize(alloced);
   	
   	// Is this necessary?    	
   	memset(&ptr[need], 0, alloced-need);
}

void BitBuf::Clear()
{
	ptr.clear();
	alloced = bitcount = 0;
}

void BitBuf::Add(unsigned value, unsigned count)
{
	Resize(bitcount + count);
	WriteBits(&ptr[0], bitcount, count, value);
	bitcount += count;
}

void BitBuf::Copy(BitBuf &src, unsigned srcbitpos, unsigned count)
{
	Resize(bitcount + count);
	CopyBits(&ptr[0], bitcount, &src.ptr[0], srcbitpos, count);
	bitcount += count;
}

BitBuf &BitBuf::Compress()
{
	BitBuf dest;
   	unsigned a, b;
   	
   	unsigned bbb, ooo, ppp;
   	unsigned bsize=UINT_MAX, Bb=0, Oo=0, Pp=0;
   	unsigned Bmax, Omax;
   	
   	bool final=false;
   	
   	for(Bmax = 1<<bbbits; (2U << Bmax) >= bitcount; --Bmax);
   	for(Omax = 1<<oobits; (2U << Omax) >= bitcount; --Omax);
   	
   	// bbb: toiston pituuden ilmaisuun kytettyjen bittien mr
   	// ooo: toiston position ilmaisuun kytettyjen bittien mr
   	// ppp: literaalin pituuden ilmaisuun kytettyjen bittien mr
	   	
   	for(bbb = bbmin; bbb < bbmin + Bmax; ++bbb)
   		for(ooo = oomin; ooo < oomin + Omax; ++ooo)
   			for(ppp = ppmin; ppp < ppmin + (1 << ppbits); ++ppp)
   			{
   				unsigned limit, blimit, olimit, plimit, poffs, pcount;
       		Again:			
       			limit = 1+bbb+ooo; /* count of bits in package record */
       			blimit = (1<<bbb)-1;
       			olimit = (1<<ooo)-1;
       			plimit = (1<<ppp)-1;
       			
       			poffs = pcount = 0;
       			
               	dest.Clear();
               	
               	dest.Add(bbb-bbmin, bbbits);
               	dest.Add(ooo-oomin, oobits);
               	dest.Add(ppp-ppmin, ppbits);
               	
               	b = bitcount;
               	
               	for(a=0; a<b; )
               	{
               		unsigned c;
               		if(a > 0)
               		{
               			unsigned blen=limit, bpos=0;
               			unsigned cmin = a>olimit ? a-olimit : 0;
               			for(c=cmin; c<a-1; ++c)
               			{
               				unsigned equlen = blen;
               				
               				while(equlen < blimit+1+limit && a+equlen < b
               				   && BitEqual(&ptr[0], a, c, equlen+1))++equlen;
               				if(equlen > blen)
               				{
               					bpos = c;
               					blen = equlen;
               				}
               			}
               			if(blen > limit)
               			{
                       		if(pcount)
                       		{
                       			if(final && verbose>4)fprintf(stderr, "   Got pp=%u at pos %u\n", pcount, poffs);
                       			dest.Add(0,      1);
                       			dest.Add(pcount, ppp);
                       			dest.Copy(*this, poffs, pcount);
                       			pcount = 0;
                       		}
               				if(final && verbose>4)
               				{
               					fprintf(stderr, "   Got bb=%u, oo=%u at pos %u\n",
               						blen-limit-1,
               						a-1-bpos,
               						a);
               				}
               				dest.Add(1,            1);
               				dest.Add(blen-limit-1, bbb);
               				dest.Add(a-1-bpos,     ooo);
               				a += blen;
               				continue;
               			}
               		}
               		
               		if(pcount == plimit)
               		{
               			if(final && verbose>4)fprintf(stderr, "   Got pp=%u at pos %u\n", pcount, poffs);
               			dest.Add(0,      1);
               			dest.Add(pcount, ppp);
               			dest.Copy(*this, poffs, pcount);
               			pcount = 0;
               		}
               		
               		if(!pcount)poffs = a;
               		++pcount;
                   	
                   	a++;
               	}
               	
           		if(pcount)
           		{
           			if(final && verbose>4)fprintf(stderr, "   Got pp=%u at pos %u\n", pcount, poffs);
           			dest.Add(0,      1);
           			dest.Add(pcount, ppp);
           			dest.Copy(*this, poffs, pcount);
           			pcount = 0;
           		}

               	if(dest.bitcount < bsize)
               	{
               		bsize = dest.bitcount;
               		Pp = ppp;
               		Bb = bbb;
               		Oo = ooo;
               	}
               	if(final)goto Done;
           	}
   	
   	if(verbose>3)
   		fprintf(stderr, 
   			"\nBest size (%5u from %5u) achieved with bbb=%u, ooo=%u, ppp=%u:\n",
   			bsize, bitcount, Bb, Oo, Pp);
   	
   	final=true;
   	bbb=Bb;
   	ooo=Oo;
   	ppp=Pp;
   	goto Again;
   	
   Done:
   	if(verbose>4)
   		fprintf(stderr, "\n");
   	
   	return *this = dest;
}

BitBuf &BitBuf::Decompress()
{
	BitBuf dest;		
	
	unsigned srcbitpos=0;
	const BYTE *const src = &ptr[0];
	
	const BYTE bbb  = bbmin+ReadBits(src, srcbitpos, bbbits); srcbitpos+=bbbits;
	const BYTE ooo  = oomin+ReadBits(src, srcbitpos, oobits); srcbitpos+=oobits;
	const BYTE ppp  = ppmin+ReadBits(src, srcbitpos, ppbits); srcbitpos+=ppbits;
	const BYTE limit = 1 + bbb + ooo;
	
	const unsigned &destbitpos = dest.bitcount;
	
//	fprintf(stderr, "  bbb=%u, ooo=%u, ppp=%u\n", bbb, ooo, ppp);

	for(; srcbitpos < bitcount; )
	{
		BYTE k = ReadBits(src, srcbitpos++, 1);
		if(k)
		{
			WORD bb = ReadBits(src, srcbitpos,       bbb);
			WORD oo = ReadBits(src, srcbitpos + bbb, ooo);
			srcbitpos += bbb + ooo;
			
		//	fprintf(stderr, "   Got bb=%u, oo=%u (destbitpos=%u).\n", bb,oo, destbitpos);
			
			bb = bb+limit+1;
			oo = destbitpos-oo-1;
			
			dest.Copy(dest, oo, bb);
		}
		else
		{
			WORD pp = ReadBits(src, srcbitpos, ppp);
			srcbitpos += ppp;
			
		//	fprintf(stderr, "   Got ppp=%u at destbitpos=%u\n", pp, destbitpos);
			
			dest.Copy(*this, srcbitpos, pp);
			srcbitpos  += pp;
		}
	}
	return *this = dest;
}

BitBuf::operator string() const
{
	return string((const char *)&ptr[0], (bitcount+7) / 8);
}

BitBuf &BitBuf::operator= (const string &s)
{
	bitcount = s.size() * 8;
	Resize(bitcount);
	memcpy(&ptr[0], s.c_str(), s.size());
	return *this;
}
